/*
 * (SLIK) SimpLIstic sKin functions
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#ifndef UI2_SCOPE_H
#define UI2_SCOPE_H


typedef enum {
	SCOPE_RENDER_POINTS = 0,
	SCOPE_RENDER_LINES
} ScopeRenderType;


typedef struct _ScopeData ScopeData;
struct _ScopeData
{
	GdkPixbuf *overlay;
	GdkPixbuf *pixbuf;
	gint width;
	gint height;
	gint x;
	gint y;

	gint horizontal;

	gint sections;
	gint *points;
	gint points_count;

	gint scale;
	ScopeRenderType mode;

	gint dynamic;
	gint stretch;
};

#define scope_set_mode(x, y) (x->mode = y)
#define scope_set_scale(x, y) (x->scale = CLAMP(y, 1, 20))


ScopeData *scope_new(GdkPixbuf *pb, gint x, gint y, gint horizontal,
		     gint w, gint h, gint dynamic, gint stretch);
ScopeData *scope_new_from_data(gchar **data, gint x, gint y, gint horizontal,
			       gint w, gint h, gint dynamic, gint stretch);
ScopeData *scope_new_from_file(const gchar *file, gint x, gint y, gint horizontal,
			       gint w, gint h, gint dynamic, gint stretch);
void scope_free(ScopeData *scope);

WidgetData *scope_register(SkinData *skin, ScopeData *scope, const gchar *key, const gchar *text_id);
WidgetData *scope_register_to_skin(const gchar *key, gchar **data, gint x, gint y,
				   gint w, gint h, gint horizontal, gint dynamic, gint stretch,
				   SkinData *skin, const gchar *text_id);

RegisterData *scope_register_key(const gchar *key, UIData *ui,
				 gint (*status_get_func)(ScopeData *scope, const gchar *key, gpointer data), gpointer status_get_data,
				 void (*mode_get_func)(ScopeData *scope, const gchar *key, gpointer data), gpointer mode_get_data);

gint scope_value_set(const gchar *key, UIData *ui, short *buffer, gint length);

gint scope_reset(const gchar *key, UIData *ui);

gint scope_set_options(const gchar *key, UIData *ui, ScopeRenderType mode, gint scale);

gint scope_widget_set(ScopeData *scope, short value);
gint scope_widget_reset(ScopeData *scope);
gint scope_widget_set_buffer(ScopeData *scope, short *buffer, gint length);

void scope_sync_to_size(ScopeData *scope, gint w, gint h, gint horizontal);


WidgetType scope_type_id(void);
void scope_type_init(void);


#endif


